package CL;

use Carp qw<croak confess>;

require Exporter;
require DynaLoader;
require AutoLoader;

@ISA = qw(Exporter DynaLoader);
$VERSION = '1.0';

## load object library
bootstrap CL $VERSION;


## subroutine used to extract constant definitions from <cwb/cl.h> and put them into hash
sub get_constant_values {
  my @hash = ();		# build list that can be used to initialise hash
  my $symbol;

  foreach $symbol (@_) {
    my $val = constant($symbol, 0);
    if ($! != 0) {		# indicates lookup failure
      croak "ERROR Constant '$symbol' not in <cwb/cl.h>";
    }
    push @hash, $symbol => $val;
  }
  return @hash;
}

## CL constants are packed into package hashes
# attribute types
%AttType = get_constant_values(
			       qw(ATT_ALIGN ATT_ALL ATT_DYN ATT_NONE ATT_POS ATT_REAL ATT_STRUC)
			       );
# argument types
%ArgType = get_constant_values(
			       qw(ATTAT_FLOAT ATTAT_INT ATTAT_NONE ATTAT_PAREF ATTAT_POS ATTAT_STRING ATTAT_VAR)
			      );

# error codes
%ErrorCode = get_constant_values(
				 qw(CDA_OK CDA_EALIGN CDA_EARGS CDA_EATTTYPE CDA_EBADREGEX CDA_EIDORNG CDA_EIDXORNG),
				 qw(CDA_ENODATA CDA_ENOMEM CDA_ENOSTRING CDA_ENULLATT CDA_ENYI CDA_EOTHER),
				 qw(CDA_EPATTERN CDA_EPOSORNG CDA_EREMOTE CDA_ESTRUC),
				);

# error symbols (indexed by <negative> error code) 
@ErrorSymbol = sort {(-$ErrorCode{$a}) <=> (-$ErrorCode{$b})} keys %ErrorCode;

# regex flags (for cl_regex2id())
%RegexFlags = (
	       '' => 0,
	       'c' => constant('IGNORE_CASE', 0), # ignore case
	       'd' => constant('IGNORE_DIAC', 0), # ignore diacritics
	       'cd' => constant('IGNORE_CASE', 0) | constant('IGNORE_DIAC', 0),	# nice short-cut trick ...
	       'dc' => constant('IGNORE_CASE', 0) | constant('IGNORE_DIAC', 0),
	      );


## CL global variables

# registry directory
$Registry = cl_standard_registry();


## CL package functions

# set CL debugging level (0=no, 1=some, 2=all debugging messages)
sub set_debug_level {
  my $lvl = shift;
  $lvl = 0 if (lc $lvl) eq "none";
  $lvl = 1 if (lc $lvl) eq "some";
  $lvl = 2 if (lc $lvl) eq "all";
  croak "Usage:  CL::set_debug_level('none' | 'some' | 'all');"
    unless $lvl =~ /^[012]$/;
  CL::cl_set_debug_level($lvl);
}

# set CL memory limit (used only by makeall so far, so no point in setting it here)
sub set_memory_limit {
  my $mb = shift;
  croak "Usage:  CL::set_memory_limit(\$megabytes);"
    unless $mb =~ /^[0-9]+$/ and $mb >= 0;
  CL::cl_set_memory_limit($mb);
}

# convert '|'-delimited string into proper (sorted) feature set value
# (if 's' or 'split' is given, splits string on whitespace; returns undef if there is a syntax error)
sub make_set {
  my $string = shift;
  my $split = shift;
  if (not defined $split) {
    $split = 0;
  }
  elsif ((lc $split) =~ /^s(plit)?$/) {
    $split = 1;
  }
  elsif ($split != 1) {
    croak "Usage:  \$feature_set = CL::make_set(\$string [, 'split' | 's']);";
  }

  return CL::cl_make_set($string, $split);
}

# 
 
# compute intersection of two feature sets (CQP's 'unify()' function)
# (returns undef if there is a syntax error)
sub set_intersection {
  croak "Usage:  \$fs = CL::set_intersection(\$set1, \$set2);"
    unless @_ == 2;
  return CL::cl_set_intersection(@_);
}

# compute cardinality of feature set (= "size", i.e. number of elements)
# (returns undef if there is a syntax error)
sub set_size {
  croak "Usage:  \$N = CL::set_size(\$feature_set)l"
    unless @_ == 1;
  my $result = CL::cl_set_size(@_);
  if ($result < 0) {
    return undef;		# error
  }
  else {
    return $result;
  }
}



#
#  ------------  CL::Attribute objects  ------------
#

package CL::Attribute;
use Carp;

sub new {
  my $class = shift;
  my $corpus = shift;	        # corpus object  (provided by CL::Corpus->attribute)
  my $name = shift;		# attribute name (provided by CL::Corpus->attribute)
  my $self = {};

  my $corpusPtr = $corpus->{'ptr'};
  my $ptr = CL::cl_new_attribute($corpusPtr, $name, $CL::AttType{'ATT_POS'});
  return undef unless defined $ptr;
  $self->{'ptr'} = $ptr;
  $self->{'corpus'} = $corpus;
  
  return bless($self, $class);
}

sub DESTROY {
  my $self = shift;

  # disabled because of buggy nature of CL interface!
  #  CL::cl_delete_attribute($self->{'ptr'});
}

sub max_cpos {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  $size = CL::cl_max_cpos($ptr);
  return (($size >= 0) ? $size : undef);
}

sub max_id {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  $size = CL::cl_max_id($ptr);
  return (($size >= 0) ? $size : undef);
}

sub str2id {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar str2id on all elements
    return map {scalar $self->str2id($_)} @_;
  }
  else {
    my $string = shift;
    confess "CL: str2id() method called with undefined value"
      unless defined $string;
    my $id = CL::cl_str2id($ptr, $string);
    return (($id >= 0) ? $id : undef);
  }
}
 
sub id2str {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar id2str on all elements
    return map {scalar $self->id2str($_)} @_;
  }
  else {
    my $id = shift;
    confess "CL: id2str() method called with undefined value"
      unless defined $id;
    my $str = CL::cl_id2str($ptr, $id);
    return ((defined $str) ? $str : undef);
  }
}

sub id2strlen {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar id2strlen on all elements
    return map {scalar $self->id2strlen($_)} @_;
  }
  else {
    my $id = shift;
    confess "CL: id2strlen() method called with undefined value"
      unless defined $id;
    my $len = CL::cl_id2strlen($ptr, $id);
    return (($len >= 0) ? $len : undef);
  }
}

sub id2freq {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar id2freq on all elements
    return map {scalar $self->id2freq($_)} @_;
  }
  else {
    my $id = shift;
    confess "CL: id2freq() method called with undefined value"
      unless defined $id;
    my $freq = CL::cl_id2freq($ptr, $id);
    return (($freq >= 0) ? $freq : undef);
  }
}

sub cpos2id {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar cpos2id on all elements
    return map {scalar $self->cpos2id($_)} @_;
  }
  else {
    my $cpos = shift;
    confess "CL: cpos2id() method called with undefined value"
      unless defined $cpos;
    my $id = CL::cl_cpos2id($ptr, $cpos);
    return (($id >= 0) ? $id : undef);
  }
}

sub cpos2str {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar cpos2str on all elements
    return map {scalar $self->cpos2str($_)} @_;
  }
  else {
    my $cpos = shift;
    confess "CL: cpos2str() method called with undefined value"
      unless defined $cpos;
    my $str = CL::cl_cpos2str($ptr, $cpos);
    return ((defined $str) ? $str : undef);
  }
}

## another hack: return feature-set attribute as hashref
sub cpos2features {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar cpos2features on all elements
    return map {scalar $self->cpos2features($_)} @_;
  }
  else {
    my $cpos = shift;
    confess "CL: cpos2features() method called with undefined value"
      unless defined $cpos;
    my $str = CL::cl_cpos2str($ptr, $cpos);
    if (defined $str) {
      confess "CL: cpos2features(): not a feature-set attribute"
	unless $str =~ s/^\|//;	# trailing '|' will be ignored by the split operator (see "man perlfunc")
      return {map {$_ => 1} split /\|/, $str}; # this is the reason why I just LOVE Perl
    }
    else {
      return undef;
    }
  }
}



sub regex2id {
  my $self = shift;
  my $ptr = $self->{'ptr'};
  my $regex = shift;
  my $flags = (@_) ? shift : '';

  croak "Usage:  \$att->regex2id(\$regex [, 'c' | 'd' | 'cd' ]);"
    unless defined $regex and $flags =~ /^(c?d?|dc)$/;

  return CL::cl_regex2id($ptr, $regex, $CL::RegexFlags{$flags});
}

sub idlist2freq {
  my $self = shift;
  my $ptr = $self->{'ptr'};
  
  return CL::cl_idlist2freq($ptr, @_);
}

sub idlist2cpos {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  return CL::cl_idlist2cpos($ptr, 1, @_); # return sorted cpos list
}




#
#  ------------  CL::AttStruc objects  ------------
#

package CL::AttStruc;
use Carp;

sub new {
  my $class = shift;
  my $corpus = shift;	        # corpus object  (provided by CL::Corpus->attribute)
  my $name = shift;		# attribute name (provided by CL::Corpus->attribute)
  my $self = {};

  my $corpusPtr = $corpus->{'ptr'};
  my $ptr = CL::cl_new_attribute($corpusPtr, $name, $CL::AttType{'ATT_STRUC'});
  return undef unless defined $ptr;
  $self->{'ptr'} = $ptr;
  $self->{'corpus'} = $corpus;

  return bless($self, $class);
}

sub DESTROY {
  my $self = shift;

  # disabled because of buggy nature of CL interface!
  #  CL::cl_delete_attribute($self->{'ptr'});
}


sub max_struc {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  my $size = CL::cl_max_struc($ptr);
  return (($size >= 0) ? $size : undef);
}

sub struc_values {
  my $self = shift;
  my $ptr = $self->{'ptr'};
  my $yesno = CL::cl_struc_values($ptr);
  return (($yesno >= 0) ? $yesno : undef);
}

sub cpos2struc {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar cpos2struc on all elements
    return map {scalar $self->cpos2struc($_)} @_;
  }
  else {
    my $cpos = shift;
    confess "CL: cpos2struc() method called with undefined value"
      unless defined $cpos;
    my $struc = CL::cl_cpos2struc($ptr, $cpos);
    return (($struc >= 0) ? $struc : undef);
  }
}

sub struc2cpos {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  croak 'Usage:  ($start, $end) = $att->struc2cpos($struc_num);'
    unless @_ == 1 and wantarray;
  my $struc = shift;
  confess "CL: struc2cpos() method called with undefined value"
    unless defined $struc;
  return CL::cl_struc2cpos($ptr, $struc);
}

sub cpos2struc2cpos {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  croak 'Usage:  ($start, $end) = $att->struc2cpos($cpos);'
    unless @_ == 1 and wantarray;
  my $cpos = shift;
  confess "CL: cpos2struc2cpos() method called with undefined value"
    unless defined $cpos;
  return CL::cl_cpos2struc2cpos($ptr, $cpos);
}

sub struc2str {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar struc2str on all elements
    return map {scalar $self->struc2str($_)} @_;
  }
  else {
    my $struc = shift;
    confess "CL: struc2str() method called with undefined value"
      unless defined $struc;
    my $str = CL::cl_struc2str($ptr, $struc);
    return ((defined $str) ? $str : undef);
  }
}

sub cpos2struc2str {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar cpos2struc2str on all elements
    return map {scalar $self->cpos2struc2str($_)} @_;
  }
  else {
    my $cpos = shift;
    confess "CL: cpos2struc2str() method called with undefined value"
      unless defined $cpos;
    my $str = CL::cl_cpos2struc2str($ptr, $cpos);
    return ((defined $str) ? $str : undef);
  }
}

# in anticipation of the new object-oriented CL interface specification,
# we define the following alias:
sub cpos2str {
  return cpos2struc2str(@_);
}

## another hack: return feature-set attribute as hashref
sub cpos2features {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar cpos2features on all elements
    return map {scalar $self->cpos2features($_)} @_;
  }
  else {
    my $cpos = shift;
    confess "CL: cpos2features() method called with undefined value"
      unless defined $cpos;
    my $str = CL::cl_cpos2struc2str($ptr, $cpos);
    if (defined $str) {
      confess "CL: cpos2features(): not a feature-set attribute"
	unless $str =~ s/^\|//;	# trailing '|' will be ignored by the split operator (see "man perlfunc")
      return {map {$_ => 1} split /\|/, $str}; # this is the reason why I just LOVE Perl
    }
    else {
      return undef;
    }
  }
}




#
#  ------------  CL::AttAlign objects  ------------
#

package CL::AttAlign;
use Carp;

sub new {
  my $class = shift;
  my $corpus = shift;	        # corpus object  (provided by CL::Corpus->attribute)
  my $name = shift;		# attribute name (provided by CL::Corpus->attribute)
  my $self = {};

  my $corpusPtr = $corpus->{'ptr'};
  my $ptr = CL::cl_new_attribute($corpusPtr, $name, $CL::AttType{'ATT_ALIGN'});
  return undef unless defined $ptr;
  $self->{'ptr'} = $ptr;
  $self->{'corpus'} = $corpus;
  
  return bless($self, $class);
}

sub DESTROY {
  my $self = shift;

  # disabled because of buggy nature of CL interface
  #  CL::cl_delete_attribute($self->{'ptr'});
}


sub max_alg {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  my $size = CL::cl_max_alg($ptr);
  return (($size >= 0) ? $size : undef);
}

sub has_extended_alignment {
  my $self = shift;
  my $ptr = $self->{'ptr'};
  my $yesno = CL::cl_has_extended_alignment($ptr);
  return (($yesno >= 0) ? $yesno : undef);
}

sub cpos2alg {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  if (wantarray) {		# list context: map scalar cpos2alg on all elements
    return map {scalar $self->cpos2alg($_)} @_;
  }
  else {
    my $cpos = shift;
    confess "CL: cpos2alg() method called with undefined value"
      unless defined $cpos;
    my $alg = CL::cl_cpos2alg($ptr, $cpos);
    return (($alg >= 0) ? $alg : undef);
  }
}

sub alg2cpos {
  my $self = shift;
  my $ptr = $self->{'ptr'};

  croak 'Usage:  ($src_start, $src_end, $trg_start, $trg_end) = $att->alg2cpos($alg_num);'
    unless @_ == 1 and wantarray;
  my $alg = shift;
  confess "CL: alg2cpos() method called with undefined value"
      unless defined $alg;
  return CL::cl_alg2cpos($ptr, $alg);
}







#
#  ------------  CL::Corpus objects  ------------
#

# $corpus = new CL::Corpus "name";
# $lemma = $corpus->attribute("lemma", 'p');     # returns CL::Attribute object (positional attribute)
# $article = $corpus->attribute("article", 's'); # returns CL::AttStruc object  (structural attribute)
# $french = $corpus->attribute("name-french", 'a');  # returns CL::AttAlign     (alignment attribute)
# undef $corpus;                                 # delete corpus from memory

package CL::Corpus;
use Carp;

sub new {
  my $class = shift;
  my $corpusname = shift;
  my $self = {};

  # try to open corpus (corpus name needs to be all lowercase)
  $corpusname =~ tr[A-Z][a-z];
  my $ptr = CL::cl_new_corpus(
			      (defined $CL::Registry) ? $CL::Registry : CL::cl_standard_registry(),
			      $corpusname
			     );
  return undef unless defined $ptr;
  $self->{'ptr'} = $ptr;
  return bless($self, $class);
}

sub DESTROY {
  my $self = shift;

  # disable because of buggy nature of CL interface
  #   CL::cl_delete_corpus($self->{'ptr'});
}

sub attribute {
  my $self = shift;
  my $name = shift;
  my $type = shift;

  if ($type eq 'p') {
    return (new CL::Attribute $self, $name);
  }
  elsif ($type eq 's') {
    return (new CL::AttStruc $self, $name);
  }
  elsif ($type eq 'a') {
    return (new CL::AttAlign $self, $name);
  }
  else {
    croak "USAGE: \$corpus->attribute(\$name, 'p' | 's' | 'a')";
  }
}




package CL;			# back to main package for autosplitter's sake
1;
__END__


=head1 NAME

CL - Perl interface to the IMS CWB/Corpus Library

=head1 SYNOPSIS

  use CL;

  print "Registry path = ", $CL::Registry, "\n";
  $CL::Registry .= ":/home/my_registry";         # add your own registry directory

  # CL::Corpus objects
  $corpus = new CL::Corpus "HANSARD-EN";         # name of corpus can be upper or lower case
  undef $corpus;                                 # delete corpus from memory

  # CL::Attribute objects
  $lemma = $corpus->attribute("lemma", 'p');     # returns CL::Attribute object (positional attribute)
  $corpus_length = $lemma->max_cpos;             # valid cpos values are 0 .. $corpus_length-1
  $lexicon_size = $lemma->max_id;                # valid id values are 0 .. $lexicon_size-1

  $id  = $lemma->str2id($string); 
  @idlist = $lemma->str2id(@strlist);            # all scalar functions map to lists in list context
  $str = $lemma->id2str($id);
  $len = $lemma->id2strlen($id);
  $f   = $lemma->id2freq($id);
  $id  = $lemma->cpos2id($cpos);
  $str = $lemma->cpos2str($cpos);

  @idlist = $lemma->regex2id($re);               # regular expression matching
  @cpos = $lemma->idlist2cpos(@idlist);	         # accessing the index
  $total_freq = $lemma->idlist2freq(@idlist);    # better check the list size first on large corpora

  undef $lemma;                                  # drop attribute from memory

  # CL::AttStruc objects
  $chapter = $corpus->attribute("chapter", 's'); # returns CL::AttStruc object  (structural attribute)
  $number_of_regions = $chapter->max_struc;      # valid region numbers are 0 .. $number_of_regions-1
  $has_values = $chapter->struc_values;          # are regions annotated with strings?

  $struc = $chapter->cpos2struc($cpos);          # returns undef if not $cpos is not in <chapter> region
  ($start, $end) = $chapter->struc2cpos($struc); # returns empty list on error -> $start is undefined
  ($start, $end) = $chapter->cpos2struc2cpos($struc);  # returns empty list if not in <chapter> region
  $str  = $chapter->struc2str($struc);           # always returns undef if not $chapter->struc_values
  $str  = $chapter->cpos2str($cpos);             # combines cpos2struc() and struc2str() 

  undef $chapter;

  # feature sets (CL::Attribute and CL::AttStruc)
  $np_f = $corpus->attribute("np_feat", 's');    # both p-attributes and s-attributes can store feature sets
  $features = $np_f->cpos2features($cpos);       # decompose feature-set attribute into hashref
  if (exists $features->{"paren"}) ... 
  $fs1 = CL::make_set(keys %$features);          # construct feature set from list or string
  $fs2 = CL::make_set("paren nogen proper", 'split');
  $fs  = CL::set_intersection($fs1, $fs2);       # intersection of feature set values
  $n   = CL::set_size($fs);                      # size of feature set

  # CL::AttAlign objects
  $french = $corpus->attribute("hansard-fr", 'a'); # returns CL::AttAlign object (alignment attribute)
  $nr_of_alignments = $french->max_alg;          # alignment block numbers are 0 .. $nr_of_alignments-1
  $extended = $french->has_extended_alignment;   # extended alignment allows gaps & crossing alignments
  
  $alg = $french->cpos2alg($cpos);               # returns undef if no alignment was found
  ($src_start, $src_end, $target_start, $target_end) 
       = $french->alg2cpos($alg);                # returns empty list on error

  undef $french;

=head1 DESCRIPTION

Sorry, there is no full description for this module, since the 
B<CWB Corpus Library>, on which B<CL.pm> is based, does not have 
complete documentation. 

All of the corpus access function provided by B<CL.pm> are subject
to change in version 4.0 of the CWB. If you want to use B<CL.pm>
anyway, have a look at the test script B<test.pl> included in the
distribution. 

=head1 COPYRIGHT

Copyright 1999-2001 Stefan Evert (IMS Stuttgart).

IMS Corpus Workbench copyright 1993-2003 IMS Stuttgart.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut

