#!/bin/env perl -w
###################################################################
###                                                             ###
###      File: /projekte/CWB/tools/perl/CQP/lib/CQP.pm          ###
###   Version:                                                  ###
###    Author: Stefan Evert                                     ###
###   Purpose: Perl interface to CQP                            ###
###   Created: Tue Mar  2 17:58:17 1999                         ###
###  Modified: Thu May 26 14:10:56 2005 (evert)                 ###
###                                                             ###
###################################################################
#
# History:
#   - asynchronous execution contributed by Andrew Sayers, 2005-05-27
#


package CQP;
$VERSION = '1.0';

# external packages used
use FileHandle;
use IPC::Open3;
use IO::Select;
use CWB;

# package global variables
our $CQP = "$CWB::CQP -c";

sub new {
  my $class = shift;		# class name
  my $self = {};		# namespace for new CQP class object
  my $options = "@_";		# CQP command-line options (use at your own risk)

  # run CQP server in the background
  my $out = $self->{'out'} = new FileHandle;
  my $in = $self->{'in'} = new FileHandle;
  my $err = $self->{'err'} = new FileHandle;
  open3($out, $in, $err, "$CQP $options");
  $out->autoflush(1);

  my $version_string; # "cqp -c" should print version on startup
  my $c;
  do {
    sysread($in, $c, 1, 0);
    $version_string .= $c;
  } while ($c ne "\n");
  chomp $version_string;
  die "ERROR: CQP backend startup failed ('$CQP $options')\n"
    unless $version_string =~ /^CQP\s+(?:\w+\s+)*([0-9]+)\.([0-9]+)(?:\.b([0-9]+))?(?:\s+(.*))?$/;
  $self->{'major_version'} = $1;
  $self->{'minor_version'} = $2;
  $self->{'beta_version'} = $3;
  $self->{'compile_date'} = $4;
  if (not (			# we need cqp-2.2.b17 or newer (for query lock)
	   $1 >= 3
	   or ($1 == 2 and $2 == 2 and $3 >= 41)
	  )) {
    die "ERROR: CQP version too old. ($version_string)\n";
  }

  # error handling
  $self->{'error_handler'} = ""; # set to subref for user-defined error handler
  $self->{'status'} = 'ok';	 # status of last executed command ('ok' or 'error')
  $self->{'error_message'} = []; # arrayref to array containing message produced by last error that occurred

  # handling CQP progress bars (see documentation of exec() function)
  $self->{'progress_handler'} = undef;

  # debugging (prints more or less everything on stdout)
  $self->{'debug'} = 0;

  # stderr's select vector (in order to check error messages)
  my $bitvec = '';
  vec($bitvec, fileno($err), 1) = 1;
  $self->{'err_vec'} = $bitvec;

  # Some variables used in asynchronous execs
  $in->blocking(0);
  $err->blocking(0);
  $self->{'error_input_select'} = new IO::Select( $self->{'err'}, $self->{'in'} );
  $self->{'error_buffer'} = '';
  $self->{'input_buffer'} = '';
  $self->{'key'} = undef;

  # CQP object setup complete
  bless($self, $class);

  # the following command will collect and ignore any output which may have been produced during startup
  $self->exec("set PrettyPrint off"); # pretty-printing should be turned off for non-interactive use

  return $self;
}

sub DESTROY {
  my $self = shift;

  my $out = $self->{'out'};
  if (defined $out) {
    $out->print("exit"); 	# exit CQP backend
    $out->close;	
  }
  my $in = $self->{'in'};
  if (defined $in) {
    $in->close;
  }
}

sub check_version {
  my $self = shift;
  my ($major, $minor, $beta) = @_;
  $minor = 0 unless defined $minor;
  $beta = 0 unless defined $beta;

  my $maj = $self->{'major_version'};
  my $min = $self->{'minor_version'};
  my $bet = $self->{'beta_version'};
  if ( $maj > $major
       or ($maj == $major and $min > $minor)
       or ($maj == $major and $min == $minor and $bet >= $beta)
     ) {
    return 1;
  }
  else {
    return 0;
  }
}

sub getline {
  our $self = shift;
  my $timeout = shift;
  my $debug = $self->{'debug'};
  my $line;

  if ($self->{'input_buffer'} =~ s/^(.*?)\n+//s) { # If there's a line of input waiting, don't read.
      $line = $1;
  } else {
      do {
	  foreach my $fh ($self->{'error_input_select'}->can_read($timeout)) {
	      if ($fh == $self->{'err'}) {
		  sysread($self->{'err'}, $self->{'error_buffer'}, 256, length($self->{'error_buffer'})) || die $!;
		  if (my $handler = $self->{'progress_handler'}) { # If no progress handler is registered, don't treat progress lines specially
		      while ($self->{'error_buffer'} =~ s/^([^\n]*?)\r//s) { # Messages ending in carriage returns are information...
			  print "CQP 2>> $1<carriage return>\n"
			      if $debug;
			  next unless $1 =~ /\[pass ([0-9]+) of ([0-9]+): *(.*)\]/;
			  my ($pass, $num_passes, $message) = ($1, $2, $3); # So weird things don't happen when the callback uses regexps
			  my $progress = ($message =~ /([0-9]+)\%\s*complete/) ? $1+0 : undef; # extract progress percentage, if present (makes callbacks easier to write)
			  $handler->($pass, $total, $progress, $message);
		      };
		  };

		  if ($self->{'error_buffer'} =~ /\n/) { # ... messages ending in newlines are errors
		      $self->{'status'} = 'error';
		      push @{$self->{'error_message'}}, split(/\n/, $self->{'error_buffer'});
		      $self->error("**** CQP ERROR ****", @{$self->{'error_message'}});
		  };
	      } else {
		  sysread($self->{'in'}, $self->{'input_buffer'}, 256, length($self->{'input_buffer'})) || die $!;
		  $self->{'input_buffer'} =~ s/^\n+//; # Remove empty lines from the beginning
	      };
	  };
      } until (defined($timeout) || $self->{'input_buffer'} =~ s/^(.*?)\n+//s); # If there's a timeout, only read once...
      return '' unless defined($line = $1); # ... and then return an empty string if there's no line to read
  };

  print "CQP >> $line\n"
      if $debug;
  $line =~ tr[\r][]d;		# ignore CR codes CQP may produce in PrettyPrint mode

  return $line unless $line eq '-::-EOL-::-';	# special line printed by ".EOL.;" marks end of command output

  print "CQP ", "-" x 60, "\n"
	  if $debug;
  return undef;

};

# execute CQP command
sub aexec {
  my $self = shift;
  my $cmd = shift;
  my $debug = $self->{'debug'};

  die "USAGE:  \$cqp->exec(\$cmd);\n"
    unless (defined $cmd) and ((not defined $output_handler) or (ref $output_handler eq 'CODE'));

  $self->{'error_buffer'} = ''; # buffer all input from STDERR into this variable
  $self->{'input_buffer'} = ''; # buffer all input from STDERR into this variable

  unless ($self->{'key'}) {	# Don't reset status if we're in the middle of a query
      $self->{'status'} = 'ok';
      $self->{'error_message'} = [];
  };
  chomp $cmd;			# make sure the user hasn't added a newline character
  $cmd =~ s/;\s*$//;		# ";" will be added when $cmd is sent to CQP
  print "CQP << $cmd; .EOL.;\n"
    if $debug;
  $self->{'out'}->print("$cmd; .EOL.;\n");
};

sub exec {
    my $self = shift;
    $self->aexec(@_);
    my @result = ();
    while (defined(my $line = $self->getline())) {
	push @result, $line;
    };
    return @result;
};

# run query in safe mode (query lock)
sub aquery {
  my $self = shift;
  my @result = ();
  my $key = int rand(1_000_000);

  $self->{'out'}->print("set QueryLock $key;\n");	# enter query lock mode without resetting buffers or sending an EOL
  $self->aexec(@_);					# query string + optional output handler
  $self->{'out'}->print("unlock $key;\n");		# exit query lock mode as we entered it
}

sub query {
    aquery(@_);
    my @result = ();
    while (defined(my $line = $self->getline())) { push @result, $line; };
    return @result;
}

# dump named query result into table of corpus positions
sub dump {
  my $self = shift;
  my $subcorpus = shift;
  my $from = shift;
  my $to = shift;
  my @matches = ();

  if ((not defined $subcorpus) or
      (defined $from and not defined $to)) {
    die "USAGE:  \$cqp->dump(\$named_query [, \$from, \$to]);\n";
  }
  $from = "" unless defined $from;
  $to = "" unless defined $to;

  $self->aexec("dump $subcorpus $from $to");
  while (defined(my $line = $self->getline())) {
      push @matches, [ split /\t/, $line ];
  };
  return @matches;
}

# undump named query result from table of corpus positions
sub undump {
  my $self = shift;
  my $subcorpus = shift;
  my $more_to_read = 1;

  my $with = "";			     # undump with target and keyword?
  my $n_el = undef; # number of anchors for each match (will be determined from first row)

  foreach my $row (@_) {
    my $row_el = @$row;
    if (not defined $n_el) {
      $n_el = $row_el;
      die "CQP: row arrays in undump table must have between 2 and 4 elements (first row has $n_el)"
	if $n_el < 2 or $n_el > 4;
      $with = "with target"
	if $n_el >= 3;
      $with .= " keyword"
	if $n_el >= 4;
    } else {
      die "CQP: all rows in undump table must have the same length (first row = $n_el, this row = $row_el)"
	unless $row_el == $n_el;
    }
  };

  # Lifted from perlipc(1)
  if  (   system('mknod',  "./undump-$$.pipe", 'p')
       && system('mkfifo', "./undump-$$.pipe") ) {
      die "mk{nod,fifo} ./undump-$$.pipe failed";
  };
  $self->aexec("undump $subcorpus $with < './undump-$$.pipe'");
  open (UNDUMP, ">./undump-$$.pipe") || die "can't write ./undump-$$.pipe: $!";
  print(UNDUMP scalar(@_), "\n");	  	 # number of matches (= remaining arguments)

  foreach my $row (@_) {
    $more_to_read = $self->getline(0) if defined($more_to_read); # Clear the input buffer if necessary
    print(UNDUMP join("\t", @$row), "\n");
  }
  close(UNDUMP);
  while (defined($more_to_read)) { $more_to_read = $self->getline }; # Clear the input buffer if necessary
  unlink("./undump-$$.pipe");

  return $self->ok;				# return success status of undump command
}


# compute frequency distribution over attribute values (single values or pairs) using group command;
# note that the arguments are specified in the logical order, in contrast to "group"
sub group {
  my $self = shift;
  if (@_ < 2 or @_ > 4) {
    die 'USAGE:  $cqp->group($named_query, "$anchor.$att" [, "$anchor.$att"] [, $cutoff]);'."\n";
  }
  my $subcorpus = shift;
  my $spec1 = shift;
  my $spec2 = (@_) ? shift : "";
  my $cutoff = 1;
  if (@_) {
    $cutoff = shift;
  }
  elsif ($spec2 =~ /^[0-9]+$/) {
    $cutoff = $spec2 + 0;
    $spec2 = "";
  }
  die "CQP:  invalid key \"$spec1\" in group() method\n"
    unless $spec1 =~ /^(match|matchend|target[0-9]?|keyword)\.([A-Za-z0-9_-]+)$/;
  $spec1 = "$1 $2";
  if ($spec2) {
    die "CQP:  invalid key \"$spec2\" in group() method\n"
      unless $spec2 =~ /^(match|matchend|target[0-9]?|keyword)\.([A-Za-z0-9_-]+)$/;
    $spec2 = "$1 $2";
  }
  my $cmd = ($spec2) ? "group $subcorpus $spec2 by $spec1 cut $cutoff" : "group $subcorpus $spec1 cut $cutoff";
  my @rows = ();
  $self->aexec($cmd);
  while (defined(my $line = $self->getline())) {
      push @rows, [ split /\t/, $line ];
  };
  return @rows;
}

# compute frequency distribution for match strings, based on sort clause
sub count {
  my $self = shift;
  if (@_ < 2 or @_ > 3) {
    die 'USAGE:  $cqp->count($named_query, $sort_clause [, $cutoff]);'."\n";
  }
  my $subcorpus = shift;
  my $sort_clause = shift;
  my $cutoff = (@_) ? shift : 1;
  my @rows = ();
  $self->aexec("count $subcorpus $sort_clause cut $cutoff");
  while (defined(my $line = $self->getline())) {
      my ($size, $first, $string) = split /\t/, $line;
      push @rows, [$size, $string, $first, $first+$size-1];
  };
  return @rows;
}

# method to read the CQP object's status and error message variables
sub status {
  my $self = shift;

  return $self->{'status'};
}

# simplified interface for checking for CQP errors
sub ok {
  my $self = shift;
  return ($self->status eq 'ok');
}

sub error_message {
  my $self = shift;
  my $aref = $self->{'error_message'};

  return @{$aref};
}

# error messages (optionally through user-defined error handler)
sub error {
  my $self = shift;

  if (ref $self->{'error_handler'} eq 'CODE') {	# subref
    &{$self->{'error_handler'}}(@_);
  }
  else {
    foreach my $line (@_) {
      print STDERR "$line\n";
    }
  }
}

# set user-defined error handler
sub set_error_handler {
  my $self = shift;
  my $handler = shift;

  die "USAGE:  \$cqp->set_error_handler(\\\&my_error_handler);\n"
    unless (not defined $handler) or (ref $handler eq 'CODE');
  $self->{'error_handler'} = $handler;
}

# set user-defined handler for CQP progress bar (automatically turns progress bar display on/off)
sub set_progress_handler {
  my $self = shift;
  my $handler = shift;

  die "USAGE:  \$cqp->set_progress_handler(\\\&my_progress_handler);\n"
    if (defined $handler) and (ref $handler ne 'CODE');
  if ($handler) {
    $self->exec("set ProgressBar on");
    $self->{'progress_handler'} = $handler;
  }
  else {
    $self->{'progress_handler'} = undef;
    $self->exec("set ProgressBar off");
  }
}

# execution-pause handler to process information from the progressbar.
# Note: makes callbacks to $self->{'progress_handler'}, with arguments ($pass, $total, $progress [0 .. 100], $message)
sub handle_progressbar {
  my $self = shift;
  my $handler = $self->{'progress_handler'};
  my $debug = $self->{'debug'};

  my $was_blocking = $self->{'err'}->blocking(0); # don't block when reading from CQP's STDERR

  my $sel_both = new IO::Select( $self->{'err'}, $self->{'in'} );
  my $sel_err = new IO::Select( $self->{'err'});
  my $err_buf = ''; # buffer all data from STDERR into this variable

  local $/;			 # turn on "slurpy" reads within this function

  while ($sel_both->can_read) {
    if ($sel_err->can_read(0)) { # process anything we get from STDERR before we look at STDOUT
      $err_buf .= $self->{'err'}->getline();	# should read as much as is currently available (because stderr is set to non-blocking)
      while ($err_buf =~ s/^([^\n]*?)\r//s) { # Messages ending in carriage returns are information...
	next unless $1 =~ /\[pass ([0-9]+) of ([0-9]+): *(.*)\]/;
	my ($pass, $num_passes, $message) = ($1, $2, $3); # So weird things don't happen when the callback uses regexps
	my $progress = ($message =~ /([0-9]+)\%\s*complete/) ? $1+0 : undef; # extract progress percentage, if present (makes callbacks easier to write)
	$handler->($pass, $total, $progress, $message);
      };
      if ($err_buf =~ /\n/) { # ... messages ending in newlines are errors
	while ($sel_err->can_read) {		# make sure we have captured the complete error message ...
	  $err_buf .= $self->{'err'}->getline();
	}
	$self->{'status'} = 'error';		# ... and throw an error
	$self->{'error_message'} = [ split(/\n/, $err_buf) ];
	$self->error("**** CQP ERROR ****", @$self->{'error_message'});
      };
    } else {
      last; # data on STDOUT indicates that CQP has finished processing the command
    };
  };
  
  $self->{'err'}->blocking($was_blocking);	# restore blocking I/O mode for STDERR stream
};

# switch debugging output on/off
sub debug {
  die "USAGE:  \$prev_status = \$cqp->debug( 1 | 0 ) ;\n"
    unless @_ == 2;
  my $self = shift;
  my $on = shift;
  my $prev = $self->{'debug'};
  $self->{'debug'} = $on;
  return $prev;
}


return 1;


__END__


=head1 NAME

  CQP - Create a CQP server process as a Perl object


=head1 SYNOPSIS

  use CQP;

  # start CQP server process in the background
  $my_cqp = new CQP;
  $my_cqp = new CQP("-r /corpora/registry", "-I /global/init.cqp");

  # check for required CQP version (
  $ok = $my_cqp->check_version($major [, $minor [, $beta]]);

  # execute CQP command and check whether error has occurred
  @lines = $my_cqp->exec($my_cmd);
  if (not $my_cqp->ok) {
    @cqp_error_message = $my_cqp->error_message;
    my_error_handler();
  }

  # it's easier to use an automatic error handler
  $my_cqp->set_error_handler(\&my_error_handler);

  # use callback function to process output of CQP command
  $my_cqp->aexec($my_cmd);
  while (defined(my $line = $cqp->getline()) {
    # process $line
  };

  # capture progress messages during query execution
  $my_cqp->set_progress_handler(\&my_progress_handler);

  # execute in query lock mode (to improve security of CGI scripts)
  @lines = $my_cqp->query($my_query);

  # dump a named query and return corpus positions of anchors as table
  @matches = $my_cqp->dump("Last" [, $from, $to]);
  foreach $m (@matches) {
    ($match, $matchend, $target, $keyword) = @{$m};
  }

  # create named query table from table of corpus positions
  $my_cqp->undump("Copy", @matches);  # produces copy of "Last"

  # compute frequency distribution with the "group" command
  @rows = $my_cqp->group("A", "match.word", "matchend.lemma");
  foreach $row (@rows) {
    ($match, $matchend, $f) = @{$row};
  }

  # compute frequencies of matching strings with the "count" command
  @rows = $my_cqp->count("A", 'by word %c', $cutoff);
  foreach $row (@rows) {
    ($f, $string, $first, $last) = @{$row};
  }

  # close down CQP server (exits gracefully)
  undef $my_cqp;


=head1 DESCRIPTION

When you create a I<CQP> object, CQP (the IMS Corpus Query Processor)
is run in the background. Every I<CQP> object has its own CQP server
process. A I<CQP> object can execute any CQP command, and returns the
output which CQP would normally print to C<STDOUT>.

Error detection is implemented by monitoring the server's C<STDERR>
stream. Error messages printed by CQP can be obtained through an
object method call.


=head1 METHODS

The following methods are available:

=over 4

=item $cqp = CQP->new;

Spawn new CQP background server. The object returned is used to
communicate with the server process.

=item $cqp = CQP->new('-r /corpora/registry');

Optional arguments of the C<new> method are passed as command-line
options to CQP. Use at your own risk.

=item $ok = $cqp->check_version($major, $minor, $beta);

Check for minimum required CQP version, i.e. 
CQP-E<lt>$majorE<gt>.E<lt>$minorE<gt>.bE<lt>$betaE<gt>
$minor and $beta may be omitted, in which case they default to 0.  
Note that the B<CQP> module automatically checks on startup that 
the CQP version is compatible with its own requirements.

=item @results = $cqp->exec($cmd);

Execute a I<single> CQP command by sending the string $cmd to CQP's
input. This call will block until CQP has completed command execution.
Regular CQP output in response to the command will be returned as an
array. Each array element corresponds to one line of CQP output (usually a
I<kwic> line) with the trailing newline stripped. CQP error messages are
reported through the I<CQP> error handling mechanism (see below). 
B<NB:> If you execute multiple CQP commands in a single call, you may confuse
the Perl/CQP interface (and subsequent commands may not work properly), so
don't.

=item $cqp->aexec($cmd);

This version of the B<exec> method sends $cmd to CQP's input and
returns immediately.  You are expected to call B<getline> (see below)
to process each line of output generated by CQP.  This mechanism is
much more memory-efficient for processing large result lists or
frequency tables than collecting all lines printed by CQP in an array
that is then returned by the B<exec> method.

=item @results = $cqp->query($cmd);

Execute $cmd in I<query lock> mode, which only allows queries to be
executed (all other commands are disabled and will raise an
exception). Otherwise this method works just like B<exec>.  The
B<query> method is provided to improve security in CGI and server
applications.

=item $cqp->aquery($cmd);

Execute $cmd in I<query lock> mode and returns immediately.  This
method behaves like B<aexec>, just as B<query> behaves like B<aexec>

=item $line = $cqp->getline([ $timeout ])

Get at most one line of output from CQP, with the trailing newline
stripped.  Returns B<undef> when the end of the responses to a command
is detected.  If I<$timeout> is specified, B<getline> retuns the empty
string '' if it hasn't received a complete line of input within that
many seconds.  Specifying a timeout of 0 will make B<getline> check
whether there a line ready immediately without blocking at all.

=item @matches = $cqp->dump($named_query [, $from, $to]);

Retrieve query matches in the form of a table of corpus positions for the
anchor points, using CQP's C<dump> command. Each element of I<@matches>
corresponds to one match and is a reference to an anonymous array with the
elements

  [match, matchend, target, keyword]

specifying the respective corpus positions.  Undefined targets and keywords
are set to -1 (especially when no targets or keywords were set for this query
result). If C<$from> and C<$to> are specified, only matches from that range
are returned. B<NB:> the line numbers of matches start from 0 and are
interpreted with respect to the current sort order of I<$named_query>.

=item $cqp->undump($named_query, @matches);

Construct a named query result from a table of corpus positions (i.e. the
opposite of the B<dump> method).  Each element of I<@matches> is a reference
to an anonymous array

  [match, matchend, target, keyword]

that represents the anchor points of a single match.  The C<target> and
C<keyword> anchors are optional, but every anonymous array in the argument
list has to have the same length.  When the matches are not sorted in
ascending order, CQP will automatically create an appropriate sort index for
the undumped query result. 

=item @rows = $cqp->group($named_query, $spec1 [, $spec2] [, $cutoff]);

Compute a frequency distribution over one or two keys given by I<$spec1> and
I<$spec2>, which must both have the form C<"I<$anchor>.I<$attribute>">
(I<$anchor> is one of C<match>, C<matchend>, C<target>, or C<keyword>, and
I<$attribute> is the name of a positional or structural attribute).  Note that
the order of the arguments is different from the counterintuitive syntax of
the CQP C<group> command.  With two keys I<$spec1> and I<$spec2>, 
it corresponds to the CQP command 

  group $named_query $anchor2 $att2 by $anchor1 $att1;

Each element of the returned array I<@rows> represents a row of the
resulting frequency table, given by an anonymous array with the values

  (value1, [value2,] frequency) 

Note that C<value2> is only present in the two-key form with I<$spec2>.  If
the optional frequency cutoff is specified, only rows with C<< frequency >=
I<$cutoff> >> are returned.

=item @rows = $cqp->count($named_query, $sort_clause [, $cutoff]);

Compute frequency counts for strings matching the query I<$named_query>.  The
I<$sort_clause> is passed directly to CQP's C<count> command and determines
the precise strings to be counted, as well as normalisation and display options.
After execution of the command, I<$named_query> is sorted in the corresponding 
order so that instances of each matching string can easily be retrieved.

Each element of the returned array I<@rows> represents a row of the resulting
frequency table, given by an anonymous array with the values

  (frequency, string, first, last)

Here, C<first> and C<last> represent the range of line numbers corresponding
to instances of C<string>.  They are no longer valid when the sort order of
I<$named_query> is changed by a subsequent C<sort> or C<count> command.  Note
that C<string> contains the values of the p-attribute selected by
I<$sort_clause>, possibly including normalisation and reverse spelling.  If
the optional frequency cutoff is specified, only rows with C<< frequency >=
I<$cutoff> >> are returned.

=item $cqp->status;

Returns the status of the last I<exec>ed command, which is either 'ok'
or 'error'.

=item $cqp->ok;

Returns C<True> iff the status of the last I<exec>ed command is 'ok'.
This method is an abbreviation for the expression

  ($cqp->status eq 'ok')

=item @lines = $cqp->error_message;

Returns the last error reported by CQP. This is not reset automatically, so
you need to check I<$cqp-E<gt>status;> in order to find out whether the error
message was actually produced by the last command.  The return value is a list
of error message lines with newlines stripped.

=item $cqp->error(@lines);

Used internally to report errors and CQP error messages. May be used
to add error conditions in the main program that are connected with a
given I<CQP> object to this object's error listing.

=item $cqp->set_error_handler(\&my_error_handler);

Set user-defined error handler callback function for this CQP process.  All
internal error messages will be sent to the callback function, which receives
a list of error message lines as arguments. When B<getline> or any the commands
that build on it detect an error, they will effectively execute

  my_error_handler($cqp->error_message);

Calling C<$cqp-E<gt>set_error_handler(undef)> will restore the default
behaviour of echoing error messages on C<STDOUT>. If you create more
than one I<CQP> object, you can use closures over anonymous
subroutines to provide each server process with its own error handler.

=item $cqp->set_progress_handler(\&my_progress_handler);

Set user-defined callback function to handle progress messages generated by
CQP during query execution (and some other functions such as C<set target> and
C<group>).  Setting the progress handler callback automatically enables
progress messages in the associated CQP process.  During query execution, the
B<CQP> module will now capture the progress messages, parse them, and invoke the
callback function 

  my_progress_handler($pass, $total, $progress, $message);

for every progress message printed by CQP.  The arguments are the current pass
(I<$pass>), the total number of passes (I<$total>), completion percentage in
the current pass (I<$progress>, an integer in the range C<0 .. 100> or
B<undef> if percentage could not be determined), and the full text of the
progress message (I<$message>).  When commands are executed that show no
progress information, the callback simply will not be invoked.  Use C<<
$cqp->set_progress_handler(undef) >> to deactivate progress messages.

B<NB:> This is an experimental feature and should be used with caution in a
production environment.  It is recommended to activate the progress handler
B<only> when executing commands that actually print progress messages.  Note
that the handler may cause lock-ups or spurious CQP errors (when some of the
progress messages are not caught by the handler, they will be misinterpreted
as error messages), although such problems have not been reported so far.

=item undef $cqp;

Exits CQP gracefully by issuing an "exit;" command.

=back


=head1 VARIABLES

It should normally not be necessary to access the internal variables
of I<CQP> objects directly. Use methods instead. However, the
following variables may be of interest to the advanced user:

=over 4

=item $CQP::CQP

The system command used to create a CQP server process.

=item $cqp->{'major_version'}

=item $cqp->{'minor_version'}

=item $cqp->{'beta_version'}

=item $cqp->{'compile_date'}

Version information provided by the CQP server process.

=back


=head1 COPYRIGHT

Copyright 1999-2005 Stefan Evert (IMS Stuttgart). Progress bar handler contributed by Andrew Sayers. 

CQP copyright 1993-2005 IMS Stuttgart.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut
