# Before `make install' is performed this script should be runnable with
# `make test'. After `make install' it should work as `perl test.pl'

BEGIN { $| = 1; print "Running tests for CWB.pm:\n"; }

# step 1: load the CWB/Perl library
END {
  if (not $loaded) {
    print "!! can't load CWB.pm (aborted)\n";
    exit 1;
  }
  if ($tempfile and -f $tempfile) {
    unlink $tempfile;		# clean up if reading / writing failed
  }
}

use CWB;
$loaded = 1;
print " - use CWB.pm ok\n";

# step 2: check that CQP and all tools are available and can be executed
print " - checking CQP and tools\n";
print "   + CQP ...           ";
try_print_version("$CWB::CQP -v");
print "   + encode ...        ";
try_print_version("$CWB::Encode -h");
print "   + makeall ...       ";
try_print_version("$CWB::Makeall -h");
print "   + decode ...        ";
try_print_version("$CWB::Decode -h");
print "   + lexdecode ...     ";
try_print_version("$CWB::Lexdecode -h");
print "   + describe-corpus...";
try_print_version("$CWB::DescribeCorpus -h");
print "   + huffcode ...      ";
try_print_version("$CWB::Huffcode -h");
print "   + compress-rdx ...  ";
try_print_version("$CWB::CompressRdx -h");
print "   + itoa ...          ";
try_print_version("$CWB::Itoa -h");
print "   + atoi ...          ";
try_print_version("$CWB::Atoi -h");
print "   + s-encode ...      ";
try_print_version("$CWB::SEncode -h");
print "   + s-decode ...      ";
try_print_version("$CWB::SDecode -h");
print "   + scan-corpus ...   ";
try_print_version("$CWB::ScanCorpus -h");

# step 3: test the OpenFile function with magical compression and decompression
$base = "/tmp/CWB.test.$$";

print " - testing .Z compression\n";
{
  local $SIG{PIPE} = "IGNORE";	# so broken pipe signal won't make test fail
  $supported = eval { try_write_read_file("$base.Z") } ;
  if (not defined $supported) {
    print "!! .Z compression not supported (skipped)\n"; 
  }
}
print " - testing .gz compression\n";
try_write_read_file("$base.gz");

# step 4: test temporary file objects, which auto-remove on exit
print " - testing CWB::TempFile (auto-removing)\n";

print "   + unnamed\n";
$tf = test_temp_file();
$tf->close;
die "ERROR: failed to auto-remove temporary file\n"
  if -f $tf->{NAME};
print "     + deleted\n";

print "   + compressed\n";
$tf = test_temp_file("MyFile.gz");
die "ERROR: temporary file removed prematurely\n"
  unless -f $tf->{NAME};
print "     + auto-remove pending\n";
undef $tf;			# delete CWB::TempFile object
die "ERROR: failed to auto-remove temporary file\n"
  if -f $tf->{NAME};
print "     + deleted\n";

# step 5: test shell command execution
print " - testing CWB::Shell::Cmd()\n";
print "   + running 'ls -l'\n";
@lines = ();
$errlvl = CWB::Shell::Cmd("ls -l", \@lines);
$N = @lines;
if ($N > 0 and $errlvl == 0 and grep {/README/} @lines) {
  print "   + ok ($N lines of output)\n";
}
else { 
  die "ERROR: error level $errlvl, $N lines of output\n";
}
print "   + running command (warning on stderr)\n";
$errlvl = CWB::Shell::Cmd("perl -e 'print STDERR \"THIS LINE SHOULD BE SHOWN AS A WARNING.\\n\"'");
if ($errlvl == 3) {
  print "\n";
  print "   + ok (error level 3)\n";
}
else {
  die "ERROR: expected error level 3, got $errlvl\n";
}

# step 6: test registry file editor
print " - testing CWB::RegistryFile\n";

print "   + loading test_registry/dickens\n";
$reg = new CWB::RegistryFile "test_registry/dickens";
die "ERROR: failed to read test registry entry\n"
  unless defined $reg;
print "   + checking corpus header\n";
print "       NAME ", $reg->name, "\n";
print "       ID   ", $reg->id, "\n";
print "       HOME ", $reg->home, "\n";
print "       INFO ", $reg->info, "\n";
die "ERROR: data incorrect\n"
  unless $reg->name =~ /Charles Dickens/ and $reg->id eq "dickens";
print "   + listing corpus properties\n";
foreach $property ($reg->list_properties) {
  print "       ##:: $property = \"", $reg->property($property), "\"\n";
}
print "   + listing attributes\n";
$N_att = 0;
foreach $type (qw<p s a>) {
  @attr = $reg->list_attributes($type);
  $N = @attr;
  printf "       %2d $type-attributes\n", $N;
  $N_att += $N;
}
@attr = $reg->list_attributes();
$N = @attr;
printf "       %2d attributes in total\n", $N;
die "ERROR: inconsistent attribute counts\n"
  unless $N == $N_att;
die "ERROR: incorrect attribute info\n"
  unless $reg->attribute("lemma") eq 'p' and $reg->attribute("np1") eq 's' and $reg->line_comment("np_h") =~ /annot/;

print "  + deleting INFO field\n";
$reg->delete_info;
print "  + modyfing ID field\n";
$reg->id("test");
$reg->line_comment(":ID", "name modified by CWB::RegistryFile module");
print "  + adding corpus property\n";
$reg->property("valid", "FALSE");
print "  + deleting attributes\n";
$reg->delete_attribute("nbc");
map {$reg->delete_attribute($_)} grep {/^(np|pp)/} $reg->list_attributes('s');
print "  + adding attributes\n";
map {$reg->add_attribute("chunk$_", 's')} ("", 1..3);
$reg->add_attribute("test-fr", 'a');
$reg->line_comment("test-fr", "one alignment per corpus pair");
$reg->add_comments("chunk", "", "some comments on this block of attributes", "   with indentation");
print "  + modyfing comments\n";
$reg->set_comments("word", "", "POSITIONAL ATTRIBUTES", "");
$reg->add_comments("p", "FURTHER COMMENTS ADDED BY CWB::RegistryFile");

$reg_file = new CWB::TempFile "./test_registry/test"; # create temporary file for modified registry entry
$reg_file->finish;
$filename = $reg_file->name;
print "  + writing modified entry to $filename\n";
$reg->write($filename);
print "  + comparing to reference file\n";
$CWB::Shell::Paranoid = -1;
system("cmp $filename test_registry/reference");
$status = CWB::Shell::Cmd("cmp $filename test_registry/reference");
die "ERROR: modified registry file differs from reference file ($status)\n"
  unless $status == 0;
$reg_file->close;				# delete temporary file

# that's it -- we've passed the test
print "\nCongratulations. All tests passed.\n";
exit 0;




# try to read version information from CWB tool and print it;
# if this fails, issue an error message and exit
sub try_print_version {
  my $cmd = shift;
  my ($filename) = split " ", $cmd;
  if (-x $filename) {
    my @out = split /\n/, `$cmd 2>\&1`;
    foreach (@out) {
      if (/(Version:\s+|\s+v)([0-9]+\.[0-9]+(\.b[0-9]+)?)/) {
	print "v$2\n";
	return;
      }
    }
  }
  print "FAILED\n";
  die "ERROR: Couldn't get version information with '$cmd'.\n";
}

# try to write / read compressed file
sub try_write_read_file {
  $tempfile = shift;		# set global variable for END{} cleanup
  print "   + writing $tempfile\n";
  my $fh = CWB::OpenFile "> $tempfile";
  my @data = (1, 42, 7, -1001);
  map {print $fh "$_\n"} @data;
  $fh->close;
  die "ERROR: couldn't write file\n"
    unless -s $tempfile >= 10;
  print "   + reading $tempfile\n";
  $fh = CWB::OpenFile "< $tempfile";
  my @lines = <$fh>;
  $fh->close;
  die "ERROR: compressed data is corrupt\n"
    if @lines+0 != @data+0 or grep {$lines[$i] != $data[$i]."\n"} (0 .. $#data);
  unlink $tempfile;
}

# test temporary file objects write / read / rewind
# (returns tempfile object, so main program can test auto-remove)
sub test_temp_file {
  print "     + creating CWB::TempFile object\n";
  my $t = new CWB::TempFile (@_);
  print "     + filename: ", $t->name, "\n";
  my $i;
  print "     + writing ...\n";
  for ($i = 0; $i < 10_000; $i++) {
    $t->write("$i\n");
  }
  $t->finish;
  print "     + filesize: ", -s $t->name, "\n";
  print "     + direct read ... \n";
  my $fh = CWB::OpenFile $t->name;
  for ($i = 0; $i < 10_000; $i++) {
    my $line = $fh->getline;
    die "ERROR: file data corrupt\n"
      unless $line eq "$i\n";
  }
  die "ERROR: file data corrput (too much data)\n"
    unless $fh->eof;
  $fh->close;
  print "     + reading ... \n";
  for ($i = 0; $i < 10_000; $i++) {
    my $line = $t->read;
    die "ERROR: file data corrupt\n"
      unless $line eq "$i\n";
  }
  die "ERROR: file data corrput (too much data)\n"
    if $t->read;
  print "     + rewinding\n";
  $t->rewind;
  print "     + reading ... \n";
  for ($i = 0; $i < 10_000; $i++) {
    my $line = $t->read;
    die "ERROR: file data corrupt\n"
      unless $line eq "$i\n";
  }
  die "ERROR: file data corrput (too much data)\n"
    if $t->read;
  return $t;
}

